/*******************************************************************************
  Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
 
  (c) Copyright 1996 - 2003 Gary Henderson (gary.henderson@ntlworld.com) and
                            Jerremy Koot (jkoot@snes9x.com)

  (c) Copyright 2002 - 2003 Matthew Kendora and
                            Brad Jorsch (anomie@users.sourceforge.net)
 

                      
  C4 x86 assembler and some C emulation code
  (c) Copyright 2000 - 2003 zsKnight (zsknight@zsnes.com),
                            _Demo_ (_demo_@zsnes.com), and
                            Nach (n-a-c-h@users.sourceforge.net)
                                          
  C4 C++ code
  (c) Copyright 2003 Brad Jorsch

  DSP-1 emulator code
  (c) Copyright 1998 - 2003 Ivar (ivar@snes9x.com), _Demo_, Gary Henderson,
                            John Weidman (jweidman@slip.net),
                            neviksti (neviksti@hotmail.com), and
                            Kris Bleakley (stinkfish@bigpond.com)
 
  DSP-2 emulator code
  (c) Copyright 2003 Kris Bleakley, John Weidman, neviksti, Matthew Kendora, and
                     Lord Nightmare (lord_nightmare@users.sourceforge.net

  OBC1 emulator code
  (c) Copyright 2001 - 2003 zsKnight, pagefault (pagefault@zsnes.com)
  Ported from x86 assembler to C by sanmaiwashi

  SPC7110 and RTC C++ emulator code
  (c) Copyright 2002 Matthew Kendora with research by
                     zsKnight, John Weidman, and Dark Force

  S-RTC C emulator code
  (c) Copyright 2001 John Weidman
  
  Super FX x86 assembler emulator code 
  (c) Copyright 1998 - 2003 zsKnight, _Demo_, and pagefault 

  Super FX C emulator code 
  (c) Copyright 1997 - 1999 Ivar and Gary Henderson.



 
  Specific ports contains the works of other authors. See headers in
  individual files.
 
  Snes9x homepage: http://www.snes9x.com
 
  Permission to use, copy, modify and distribute Snes9x in both binary and
  source form, for non-commercial purposes, is hereby granted without fee,
  providing that this license information and copyright notice appear with
  all copies and any derived work.
 
  This software is provided 'as-is', without any express or implied
  warranty. In no event shall the authors be held liable for any damages
  arising from the use of this software.
 
  Snes9x is freeware for PERSONAL USE only. Commercial users should
  seek permission of the copyright holders first. Commercial use includes
  charging money for Snes9x or software derived from Snes9x.
 
  The copyright holders request that bug fixes and improvements to the code
  should be forwarded to them so everyone can benefit from the modifications
  in future versions.
 
  Super NES and Super Nintendo Entertainment System are trademarks of
  Nintendo Co., Limited and its subsidiary companies.
*******************************************************************************/

#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>
#include <signal.h>

#include <gx.h>
#include <gu.h>

#include "snes9x.h"
#include "memmap.h"
#include "debug.h"
#include "ppu.h"
#include "snapshot.h"
#include "gfx.h"
#include "display.h"
#include "apu.h"

extern uint32 joypads [5];

typedef enum { RS_READY, RS_WAIT_FOR_SYNC, RS_WAIT_FOR_VSYNC } READYSTATE;
static int gfd = -1;
static gx_memory_t tex;
static volatile READYSTATE readyState = RS_READY;
static GXINIT gxi;
static Mtx projectionMatrix,modelViewMatrix;

#define TEX_WIDTH 256
#define TEX_HEIGHT 256
#define WIDTH 640
#define HEIGHT gxi.efbHeight

#define SNES_WIDTH 256
#define SNES_HEIGHT 224

#define COPY_TOKEN 0x1F7E

#define FREE_BUFFER(a) { if (a) { free(a); a = NULL; } }
#define CLOSE(a) { if (a >= 0) { close(a); a = -1; } }

#define RGB16(r,g,b) ( (((r)/8) << 11) | (((g)/4) << 5) | ((b)/8) )
#define RGB32(r,g,b) ( ((r) << 16) | ((g) << 8) | (b) )
#define RGB32_R(clr) (((clr) >> 16) & 0xFF)
#define RGB32_G(clr) (((clr) >> 8 ) & 0xFF)
#define RGB32_B(clr) (((clr)      ) & 0xFF)

static void DrawDoneHandler()
{  
  /* now copy to fb */
  GX_CopyDisp(GX_TRUE);
  GX_Flush();
  /* set a token so we know when the copy is finished */
  GX_SetDrawSync(COPY_TOKEN);
  GX_Flush();
}

static void DrawSyncHandler(u16 token)
{
  if (token == COPY_TOKEN)
  {
    readyState = RS_WAIT_FOR_VSYNC;

    GX_Flip();
  }
}

static void VTraceHandler()
{
  /* sync completed */
  readyState = RS_READY;
}

void S9xDeinitDisplay ()
{
  if (gfd >= 0)
  {
    GX_FreeContiguousMemory(&tex);
    GX_Destroy();
    CLOSE(gfd);
  }
  
  FREE_BUFFER(GFX.Screen);
  FREE_BUFFER(GFX.SubScreen);
  FREE_BUFFER(GFX.ZBuffer);
  FREE_BUFFER(GFX.SubZBuffer);
}

static void sigHandler(int sig)
{
  printf("Received signal %d, quitting\n",sig);
  S9xExit();
}

void S9xInitDisplay (int, char **)
{
  GXTexObj texobj;
  static GXColor black = {0,0,0,0xFF};

  // install my signal handlers
  signal(SIGQUIT,sigHandler);
  signal(SIGTERM,sigHandler);
  signal(SIGINT,sigHandler);

  if (gfd < 0)
  {
    if ((gfd = open("/dev/fb0",O_SYNC | O_RDWR)) < 0)
    {
      fprintf(stderr,"Cannot open /dev/fb0\n");
      S9xExit();
    }
    
    if (GX_Init(gfd,&gxi) == NULL)
    {
      fprintf(stderr,"Cannot initialize GX\n");
      CLOSE(gfd);
      S9xExit();
    }
    
    /* allocate memory for our texture */
    tex.length = TEX_WIDTH * TEX_HEIGHT * 2;
    if (tex.user_mem)
    {
      printf("WARNING - allocating physical memory when we already have some\n");
    }
    if (GX_AllocContiguousMemory(&tex))
    {
      fprintf(stderr,"Failed to allocate physical memory\n");
      GX_Destroy();
      CLOSE(gfd);
      S9xExit();
    }  

    /* everything worked, setup the GX */
    GX_SetCopyClear(black,GX_MAX_Z24);
    
    GX_SetPixelFmt(GX_PF_RGB8_Z24, GX_ZC_LINEAR);
    
    GX_SetCullMode(GX_CULL_NONE);
    GX_SetZMode(GX_FALSE,GX_ALWAYS,GX_TRUE);
    GX_SetColorUpdate(GX_TRUE);
    GX_CopyDisp(GX_TRUE);
    GX_SetDispCopyGamma(GX_GM_1_0);
    
    GX_ClearVtxDesc();
    GX_SetVtxAttrFmt(GX_VTXFMT0,GX_VA_POS,GX_POS_XYZ,GX_F32,0);
    GX_SetVtxAttrFmt(GX_VTXFMT0,GX_VA_TEX0,GX_TEX_ST,GX_F32,0);
    
    GX_SetVtxDesc(GX_VA_POS,GX_DIRECT);
    GX_SetVtxDesc(GX_VA_TEX0,GX_DIRECT);
    
    GX_SetNumChans(0);		/* default, color = vertex color */
    GX_SetNumTexGens(1);
    GX_SetTexCoordGen(GX_TEXCOORD0,GX_TG_MTX2x4,GX_TG_TEX0,GX_IDENTITY);
    GX_SetTevOrder(GX_TEVSTAGE0,GX_TEXCOORD0,GX_TEXMAP0,GX_COLORNULL);
    GX_SetTevOp(GX_TEVSTAGE0,GX_REPLACE);
    
    GX_InitTexObj(&texobj,tex.phys_mem,TEX_WIDTH,TEX_HEIGHT,
		  GX_TF_RGB565,GX_REPEAT,GX_REPEAT,GX_FALSE);
    GX_LoadTexObj(&texobj,GX_TEXMAP0);
    
    //GX_InitTexObjLOD(&texture.texobj,GX_NEAR,GX_NEAR,0,0,0,GX_FALSE,GX_DISABLE,GX_ANISO_1);
    
    GX_InvalidateTexAll();
    
    /* load projection matrix */
    guOrtho(projectionMatrix,0,WIDTH,HEIGHT,0,-1,1); 
    GX_LoadProjectionMtx(projectionMatrix,GX_ORTHOGRAPHIC);
    
    /* load model view matrix */
    c_guMtxScale(modelViewMatrix,WIDTH,HEIGHT,1);
    GX_LoadPosMtxImm(modelViewMatrix,GX_PNMTX0);
    
    GX_SetDrawDoneCallback(DrawDoneHandler);
    GX_SetDrawSyncCallback(DrawSyncHandler);
    GX_SetVTraceCallback(VTraceHandler);
  }
  /* -------------------------------------
   *
   */
  
  Settings.SixteenBit = TRUE;
  if (Settings.ForceNoTransparency) {
    Settings.Transparency = FALSE;
  } else {
    Settings.Transparency = TRUE;
  }
 
  S9xSetRenderPixelFormat (RGB565); 

  GFX.Pitch = IMAGE_WIDTH * sizeof(uint16);
  GFX.Screen = (uint8*)malloc(GFX.Pitch * IMAGE_HEIGHT);
  GFX.SubScreen = (uint8 *) malloc(GFX.Pitch * IMAGE_HEIGHT);
  GFX.ZBuffer = (uint8 *) malloc(IMAGE_WIDTH * IMAGE_HEIGHT);
  GFX.SubZBuffer = (uint8 *) malloc(IMAGE_WIDTH * IMAGE_HEIGHT);
  GFX.Delta = (GFX.SubScreen - GFX.Screen) >> 1;
}

#define PRINTF_FPS
#ifdef PRINTF_FPS
static void subtime(struct timeval *ret,struct timeval *left,struct timeval *right)
{
  ret->tv_sec = left->tv_sec - right->tv_sec;
  ret->tv_usec= left->tv_usec- right->tv_usec;
  while (ret->tv_usec < 0)
  {
    ret->tv_usec += 1000000;
    ret->tv_sec--;
  }
}
#endif

void S9xPutImage (int width, int height)
{
  // copy from GFX.Screen to GX
  const float gs = 1.0;
  const float gt = (float)SNES_HEIGHT/(float)TEX_HEIGHT;
  int row,x;
#ifdef PRINTF_FPS
  static struct timeval last;
  static unsigned int drawnFrames,passedFrames;
  struct timeval now,elapsed;
  double fps[2];
#endif

#ifdef PRINTF_FPS
  ++passedFrames;
  gettimeofday(&now,0);
  subtime(&elapsed,&now,&last);
  if (elapsed.tv_sec >= 1)
  {
#define MICROS(a) ((a.tv_sec * 1000000) + a.tv_usec)
    fps[0] = (double)passedFrames / ((double)MICROS(elapsed) / 1000000.0);
    fps[1] = (double)drawnFrames  / ((double)MICROS(elapsed) / 1000000.0);
    printf("FPS is %f : %f\n",fps[0],fps[1]);
    
    passedFrames = 0;
    drawnFrames = 0;
    memcpy(&last,&now,sizeof(last));
  }
#endif
  /* check if we can draw or not 
  while (readyState != RS_READY)
  {
    // wait until the previous frame is done drawing
    pause();
    } */
#ifdef PRINTF_FPS
  ++drawnFrames;
#endif
  /* set flag */
  readyState = RS_WAIT_FOR_SYNC;
  /* update the texture */
  u32 *buf = (u32*)GFX.Screen;
  u32 *base = (u32*)tex.user_mem;
  for (row=0;row<SNES_HEIGHT/4;++row)
  {
    for (x=0;x<SNES_WIDTH/4;++x)
    {
      *(base++) = buf[0];
      *(base++) = buf[1];
      
      *(base++) = buf[1*SNES_WIDTH/2];
      *(base++) = buf[1*SNES_WIDTH/2+1];
      
      *(base++) = buf[2*SNES_WIDTH/2];
      *(base++) = buf[2*SNES_WIDTH/2+1];
      
      *(base++) = buf[3*SNES_WIDTH/2];
      *(base++) = buf[3*SNES_WIDTH/2+1];
      /* shift to the next block */
      buf += 2; 
    }
    /* now shift down three rows */
    buf += (3*SNES_WIDTH)/2;
  }
  dcache_flush(tex.user_mem,SNES_WIDTH * SNES_HEIGHT * 2);

  /* setup GX */
  GX_InvalidateTexAll(); 

  // ok render the triangle now 
  GX_Begin(GX_QUADS,GX_VTXFMT0,4);
  {
    GX_Position3f32(0,0,0);
    GX_TexCoord2f32(0,0);
    
    GX_Position3f32(0,1,0);
    GX_TexCoord2f32(0,gt);
    
    GX_Position3f32(1,1,0);
    GX_TexCoord2f32(gs,gt);
    
    GX_Position3f32(1,0,0);
    GX_TexCoord2f32(gs,0);
  }
  GX_End();
  
  // set draw done, my handler will be called when the GX finishes 
  GX_SetDrawDone();
  // wait until the frame is displayed
  while (readyState != RS_READY)
  {
    pause();
  }
}

void S9xSetTitle (const char *string)
{
}

bool8 S9xReadMousePosition (int which1, int &x, int &y, uint32 &buttons)
{
  return (FALSE);
}

bool8 S9xReadSuperScopePosition (int &x, int &y, uint32 &buttons)
{
  return (FALSE);
}

void S9xProcessEvents (bool8 block)
{
#define QUIT_MASK (SNES_A_MASK | SNES_B_MASK | SNES_X_MASK | SNES_Y_MASK | SNES_SELECT_MASK)
  if ( ((joypads[0] & QUIT_MASK) == QUIT_MASK) ||
       ((joypads[1] & QUIT_MASK) == QUIT_MASK) ) 
  {
    printf("You pressed the magic combo, exiting\n");
    S9xExit();
  }
}

void S9xSetPalette ()
{
}

const char *S9xSelectFilename (const char *def, const char *dir1,
			       const char *ext1, const char *title)
{
  static char path [PATH_MAX];
  char buffer [PATH_MAX];
  
  printf ("\n%s (default: %s): ", title, def);
  fflush (stdout);
  if (fgets (buffer, sizeof (buffer) - 1, stdin))
  {
    char *p = buffer;
    while (isspace (*p))
      p++;
    if (!*p)
    {
      strcpy (buffer, def);
      p = buffer;
    }

    char *q = strrchr (p, '\n');
    if (q)
      *q = 0;

    char fname [PATH_MAX];
    char drive [_MAX_DRIVE];
    char dir [_MAX_DIR];
    char ext [_MAX_EXT];

    _splitpath (p, drive, dir, fname, ext);
    _makepath (path, drive, *dir ? dir : dir1, fname, *ext ? ext : ext1);
    return (path);
  }

  return (NULL);
}

void S9xTextMode ()
{
}

void S9xGraphicsMode ()
{
}

void S9xParseDisplayArg (char **argv, int &ind, int)
{
  /*if (strcasecmp (argv [ind], "-scale2x") == 0)
    scale2x = TRUE;
    else if (strcasecmp (argv [ind], "-scale2x_adv") == 0)
    scale2x_adv=TRUE;
    else
    S9xUsage (); */
}

void S9xExtraUsage ()
{
  /*
    printf ("-scale2x      Simple 2x scaling of the output\n\n");
    printf ("-scale2x_adv    Advanced 2x scaling\n\n");  
  */
}

int S9xMinCommandLineArgs ()
{
  return (2);
}

void S9xMessage (int type, int number, const char *message)
{
  fprintf (stdout, "%s\n", message);
}


bool JustifierOffscreen()
{
  return (FALSE);
}

void JustifierButtons(uint32& justifiers)
{
#if 0
  if(IPPU.Controller==SNES_JUSTIFIER_2)
  {
    if((GUI.mouse_buttons&1)||(GUI.mouse_buttons&2))
    {
      justifiers|=0x00200;
    }
    if(GUI.mouse_buttons&4)
    {
      justifiers|=0x00800;
    }
  }
  else
  {
    if((GUI.mouse_buttons&1)||(GUI.mouse_buttons&2))
    {
      justifiers|=0x00100;
    }
    if(GUI.mouse_buttons&4)
    {
      justifiers|=0x00400;
    }
  }
#endif
}

